package ru.dbotthepony.kstarbound.defs.item

import com.google.gson.Gson
import com.google.gson.TypeAdapter
import com.google.gson.TypeAdapterFactory
import com.google.gson.reflect.TypeToken
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonToken
import com.google.gson.stream.JsonWriter
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.defs.IThingWithDescription
import ru.dbotthepony.kstarbound.defs.image.SpriteReference
import ru.dbotthepony.kstarbound.io.json.builder.FactoryAdapter
import ru.dbotthepony.kstarbound.io.json.builder.JsonImplementation
import ru.dbotthepony.kstarbound.io.json.ifString
import ru.dbotthepony.kstarbound.util.AssetPathStack

interface IItemDefinition : IThingWithDescription {
	/**
	 * Внутреннее имя предмета (ID).
	 * Не путать с именем предмета!
	 *
	 * @see shortdescription
	 * @see description
	 */
	val itemName: String

	/**
	 * Цена в пикселях
	 */
	val price: Long

	/**
	 * Редкость как [ItemRarity]
	 */
	val rarity: ItemRarity

	/**
	 * Категория предмета, определяет, в какую вкладку инвентаря оно попадает
	 */
	val category: String?

	/**
	 * Иконка в инвентаре, относительный и абсолютный пути
	 */
	val inventoryIcon: List<IInventoryIcon>?

	@JsonImplementation(InventoryIcon::class)
	interface IInventoryIcon {
		val image: SpriteReference
	}

	data class InventoryIcon(
		override val image: SpriteReference
	) : IInventoryIcon {
		class Factory(val remapper: AssetPathStack, val spriteRegistry: SpriteReference.Adapter) : TypeAdapterFactory {
			override fun <T : Any?> create(gson: Gson, type: TypeToken<T>): TypeAdapter<T>? {
				if (type.rawType == InventoryIcon::class.java) {
					return object : TypeAdapter<InventoryIcon>() {
						private val adapter = FactoryAdapter.Builder(InventoryIcon::class, InventoryIcon::image).build(gson)

						override fun write(out: JsonWriter, value: InventoryIcon?) {
							if (value == null)
								out.nullValue()
							else
								adapter.write(out, value)
						}

						override fun read(`in`: JsonReader): InventoryIcon? {
							if (`in`.peek() == JsonToken.NULL)
								return null

							if (`in`.peek() == JsonToken.STRING) {
								return InventoryIcon(spriteRegistry.parse(remapper.remap(`in`.nextString())))
							}

							return adapter.read(`in`)
						}
					} as TypeAdapter<T>
				}

				return null
			}
		}
	}

	/**
	 * Теги предмета
	 */
	val itemTags: List<String>

	/**
	 * При подборе предмета мгновенно заставляет игрока изучить эти рецепты крафта
	 */
	val learnBlueprintsOnPickup: List<String>

	/**
	 * Максимальное количество предмета в стопке
	 */
	val maxStack: Long

	/**
	 * snip
	 */
	val eventCategory: String?

	/**
	 * Заставляет предмет "использовать" сразу же при подборе
	 */
	val consumeOnPickup: Boolean

	/**
	 * Запускает следующие квест(ы) при подборе
	 */
	val pickupQuestTemplates: List<String>

	/**
	 * тип tooltip'а
	 *
	 * ссылается на конфиг окон
	 */
	val tooltipKind: String

	/**
	 * Занимает ли предмет обе руки
	 */
	val twoHanded: Boolean

	/**
	 * Заставляет SAIL/прочих болтать при подборе предмета в первый раз
	 */
	val radioMessagesOnPickup: List<String>

	/**
	 * Топливо корабля
	 */
	val fuelAmount: Long?
}
